/*
* Copyright (c) 2007, Jing Wang
* Copyright (c) 2007, Jie Liang
* Copyright (c) 2007, Multimedia Communication Laboratory, Simon Fraser University, Canada
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

#include "opj_includes.h"

#define NO_SPLIT     0
#define SPLIT_HOR    1  //horizontal slice: height is specified by slice_w
#define SPLIT_VER    2  //vertical slice: width is specified by slice_w
#define SPLIT_CB_U   3
#define SPLIT_CB_L   4

#define NO_PRED      0
#define PRED_HOR     1
#define PRED_VER     2
#define PRED_DL      3
#define PRED_DR      4
#define PRED_CROSS   5
#define PRED_BWD     6

#define LL_SPLIT    SPLIT_HOR
#define HL_SPLIT    SPLIT_HOR
#define LH_SPLIT    SPLIT_HOR
#define HH_SPLIT    SPLIT_HOR

#define LL_PRED     PRED_VER
#define HL_PRED     NO_PRED
#define LH_PRED     NO_PRED
#define HH_PRED     NO_PRED

#define LL_SUBBAND  0
#define HL_SUBBAND  1
#define LH_SUBBAND  2
#define HH_SUBBAND  3

/* <summary>                                               */
/* partition of DWT coefficients for one subband           */
/* </summary>                                              */
void mdc_split_sb(int * src, int * dst, int rw, int rh, int w, int h, int opt, int slice_w, int alter, int sb_idx)
{
	int x, y;
	int rw1,rh1;
	int rhs;
	int *dst_0;
	int *dst_1;

	switch(opt)
	{
	case SPLIT_HOR:
		rw1 = rw;
		rh1 = rh / 2;
		rhs = rh / 2 / slice_w * slice_w;
		if (alter)
		{
			dst_0 = dst + w*rh1;
			dst_1 = dst;
		}
		else
		{
			dst_0 = dst;
			dst_1 = dst + w*rh1;
		}
		for (y = 0; y < rhs; y += slice_w)
		{
			for (x = 0; x < slice_w; x ++)
			{
				memcpy(dst_0, src, sizeof(int)*rw1);
				src += w;
				dst_0 += w;
			}
			for (x = 0; x < slice_w; x ++)
			{
				memcpy(dst_1, src, sizeof(int)*rw1);
				src += w;
				dst_1 += w;
			}
		}

		// split remaining coefficients into half-half
		for (x = 0; x < rh1 - y; x ++)
		{
			memcpy(dst_0, src, sizeof(int)*rw1);
			src += w;
			dst_0 += w;
		}
		for (x = 0; x < rh1 - y; x ++)
		{
			memcpy(dst_1, src, sizeof(int)*rw1);
			src += w;
			dst_1 += w;
		}
		break;
	case SPLIT_VER:
		rw1 = rw / 2;
		rh1 = rh;
		if (alter)
		{
			dst_0 = dst + rw1;
			dst_1 = dst;
		}
		else
		{
			dst_0 = dst;
			dst_1 = dst + rw1;
		}
		for (y = 0; y < rh1; y ++)
		{
			for (x = 0; x < rw1; x += slice_w)
			{
				memcpy(dst_0 + x, src + (x<<1),           sizeof(int)*slice_w);
				memcpy(dst_1 + x, src + (x<<1) + slice_w, sizeof(int)*slice_w);
			}
			src += w;
			dst_0 += w;
			dst_1 += w;
		}

		break;
	case SPLIT_CB_U:
		rw1 = rw;
		rh1 = rh / 2;
		if (alter)
		{
			dst_0 = dst + w*rh1;
			dst_1 = dst;
		}
		else
		{
			dst_0 = dst;
			dst_1 = dst + w*rh1;
		}

		for (y = 0; y < rh / slice_w; y += 2)
		{
			for (x = 0; x < rw / slice_w; x ++)
			{
				int i, j;

				if((x+y) % 2 == 0)
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst_0 + j + x * slice_w + (y/2 * slice_w + i) * w) = *(src + j + x * slice_w + (y * slice_w + i) * w);
							*(dst_1 + j + x * slice_w + (y/2 * slice_w + i) * w) = *(src + j + x * slice_w + ((y + 1) * slice_w + i) * w);
						}
					}
				}
				else
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst_1 + j + x * slice_w + (y/2 * slice_w + i) * w) = *(src + j + x * slice_w + (y * slice_w + i) * w);
							*(dst_0 + j + x * slice_w + (y/2 * slice_w + i) * w) = *(src + j + x * slice_w + ((y + 1) * slice_w + i) * w);
						}
					}
				}
			}
		}
		break;
	case SPLIT_CB_L:
		rw1 = rw / 2;
		rh1 = rh;
		if (alter)
		{
			dst_0 = dst + rw1;
			dst_1 = dst;
		}
		else
		{
			dst_0 = dst;
			dst_1 = dst + rw1;
		}

		for (y = 0; y < rh / slice_w; y ++)
		{
			for (x = 0; x < rw / slice_w; x += 2)
			{
				int i, j;

				if((x+y) % 2 == 0)
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst_0 + j + x/2 * slice_w + (y * slice_w + i) * w) = *(src + j + x * slice_w + (y * slice_w + i) * w);
							*(dst_1 + j + x/2 * slice_w + (y * slice_w + i) * w) = *(src + j + (x + 1) * slice_w + (y * slice_w + i) * w);
						}
					}
				}
				else
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst_1 + j + x/2 * slice_w + (y * slice_w + i) * w) = *(src + j + x * slice_w + (y * slice_w + i) * w);
							*(dst_0 + j + x/2 * slice_w + (y * slice_w + i) * w) = *(src + j + (x + 1) * slice_w + (y * slice_w + i) * w);
						}
					}
				}
			}
		}
		break;
	default:  
		// simply copy the original data (NO_SPLIT)
		for (y = 0; y < rh; y ++)
		{
			memcpy(dst, src, sizeof(int)*rw);
			src += w;
			dst += w;
		}
		break;
	}
}

/* <summary>                                                              */
/* reset DWT coefficients of one subband to the original partition        */
/* </summary>                                                             */
void mdc_reset_sb(int * src, int * dst, int rw, int rh, int w, int h, int opt, int slice_w, int alter)
{
	int x, y;
	int rw1,rh1;
	int rhs;
	int *src_0;
	int *src_1;
	int *dst_0;
	int *dst_1;

	switch(opt)
	{
	case SPLIT_HOR:
		rw1 = rw;
		rh1 = rh / 2;
		rhs = rh / 2 / slice_w * slice_w;
		if (alter)
		{
			src_0 = src + w * rh1;
			src_1 = src;
		}
		else
		{
			src_0 = src;
			src_1 = src + w * rh1;
		}
		for (y = 0; y < rhs; y += slice_w)
		{
			for (x = 0; x < slice_w; x ++)
			{
				memcpy(dst, src_0, sizeof(int)*rw1);
				dst += w;
				src_0 += w;
			}
			for (x = 0; x < slice_w; x ++)
			{
				memcpy(dst, src_1, sizeof(int)*rw1);
				dst += w;
				src_1 += w;
			}
		}

		// regroup remaining coefficients
		for (x = 0; x < rh1 - y; x ++)
		{
			memcpy(dst, src_0, sizeof(int)*rw1);
			dst += w;
			src_0 += w;
		}
		for (x = 0; x < rh1 - y; x ++)
		{
			memcpy(dst, src_1, sizeof(int)*rw1);
			dst += w;
			src_1 += w;
		}
		break;
	case SPLIT_VER:
		rw1 = rw / 2;
		rh1 = rh;
		if (alter)
		{
			src_0 = src + rw1;
			src_1 = src;
		}
		else
		{
			src_0 = src;
			src_1 = src + rw1;
		}
		for (y = 0; y < rh1; y ++)
		{
			for (x = 0; x < rw1; x += slice_w)
			{
				memcpy(dst + (x<<1),           src_0 + x, sizeof(int)*slice_w);
				memcpy(dst + (x<<1) + slice_w, src_1 + x, sizeof(int)*slice_w);
			}
			dst += w;
			src_0 += w;
			src_1 += w;
		}
		break;
	case SPLIT_CB_U:
		rw1 = rw;
		rh1 = rh / 2;
		if (alter)
		{
			src_0 = src + w*rh1;
			src_1 = src;
			dst_0 = dst + w * slice_w;
			dst_1 = dst;
		}
		else
		{
			src_0 = src;
			src_1 = src + w*rh1;
			dst_0 = dst;
			dst_1 = dst + w * slice_w;
		}

		for (y = 0; y < rh / slice_w; y += 2)
		{
			for (x = 0; x < rw / slice_w; x ++)
			{
				int i, j;

				if(x % 2 == 0)
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst + j + x * slice_w + ( y      * slice_w + i) * w) = *(src_0 + j + x * slice_w + (y/2 * slice_w + i) * w);
							*(dst + j + x * slice_w + ((y + 1) * slice_w + i) * w) = *(src_1 + j + x * slice_w + (y/2 * slice_w + i) * w);
						}
					}
				}
				else
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst + j + x * slice_w + ( y      * slice_w + i) * w) = *(src_1 + j + x * slice_w + (y/2 * slice_w + i) * w);
							*(dst + j + x * slice_w + ((y + 1) * slice_w + i) * w) = *(src_0 + j + x * slice_w + (y/2 * slice_w + i) * w);
						}
					}
				}
			}
		}
		break;
	case SPLIT_CB_L:
		rw1 = rw / 2;
		rh1 = rh;
		if (alter)
		{
			src_0 = src + rw1;
			src_1 = src;
			dst_0 = dst + 1;
			dst_1 = dst;
		}
		else
		{
			src_0 = src;
			src_1 = src + rw1;
			dst_0 = dst;
			dst_1 = dst + 1;
		}

		for (y = 0; y < rh / slice_w; y ++)
		{
			for (x = 0; x < rw / slice_w; x += 2)
			{
				int i, j;

				if(y % 2 == 0)
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst + j +  x      * slice_w + (y * slice_w + i) * w) = *(src_0 + j + x/2 * slice_w + (y * slice_w + i) * w);
							*(dst + j + (x + 1) * slice_w + (y * slice_w + i) * w) = *(src_1 + j + x/2 * slice_w + (y * slice_w + i) * w);
						}
					}
				}
				else
				{
					for (i = 0; i < slice_w; i ++)
					{
						for (j = 0; j < slice_w; j ++)
						{
							*(dst + j + (x + 1) * slice_w + (y * slice_w + i) * w) = *(src_0 + j + x/2 * slice_w + (y * slice_w + i) * w);
							*(dst + j +  x      * slice_w + (y * slice_w + i) * w) = *(src_1 + j + x/2 * slice_w + (y * slice_w + i) * w);
						}
					}
				}
			}
		}
		break;
	default:  
		// simply copy the original data (NO_SPLIT)
		for (y = 0; y < rh; y ++)
		{
			memcpy(dst, src, sizeof(int)*rw);
			src += w;
			dst += w;
		}
		break;
	}
}

/* <summary>                                               */
/* partition of DWT coefficients for one tile              */
/* </summary>                                              */
void mdc_split_tile(opj_tcd_tilecomp_t * tilec, int *src, int *dst)
{
	int i;
	int *a = NULL;
	int *p = NULL;
	int w, l, h;
	int rw;			// width of the current resolution level computed                                                           */
	int rh;			// height of the current resolution level computed                                                          */
	int alter;
	int slice_w;

	w = tilec->x1-tilec->x0;
	h = tilec->y1-tilec->y0;
	l = tilec->numresolutions-1;

	// LL subband (upleft:   (0,rw) ~ (0, rh))
	rw = tilec->resolutions[0].x1 - tilec->resolutions[0].x0;
	rh = tilec->resolutions[0].y1 - tilec->resolutions[0].y0;
	a = src;
	p =	dst;
	alter = 0;
	slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : 1;
	slice_w = 1;
	mdc_split_sb(a, p, rw, rh, w, h, LL_SPLIT, slice_w, alter, LL_SUBBAND);

	for (i = 1; i <= l; i++) {

		rw = tilec->resolutions[i - 1].x1 - tilec->resolutions[i - 1].x0;
		rh = tilec->resolutions[i - 1].y1 - tilec->resolutions[i - 1].y0;

		alter = ALTER_SPLIT ? (i % 2) : 0;
		slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : rw/2 << (i - 1);

		// HL subband (upright:   (0,rw) ~ (rh,2*rw-1))
		a = src + rw;
		p =	dst + rw;
		mdc_split_sb(a, p, rw, rh, w, h, HL_SPLIT, slice_w, alter, HL_SUBBAND);

		// LH subband (downleft:  (rh,0) ~ (2*rh-1,rw))
		a = src + w * rh;
		p =	tilec->mdc_data[0] + w * rh;
		mdc_split_sb(a, p, rw, rh, w, h, LH_SPLIT, slice_w, 1-alter, LH_SUBBAND);

		// HH subband (downright: (rh,rw) ~ (2*rh-1,2*rw-1))
		a = src + w * rh + rw;
		p =	dst + w * rh + rw;
		mdc_split_sb(a, p, rw, rh, w, h, HH_SPLIT, slice_w, 1-alter, HH_SUBBAND);
	}
}

/* <summary>                                                                */
/* reset DWT coefficients of one tile to the original partition             */
/* </summary>                                                               */
void mdc_reset_tile(opj_tcd_tilecomp_t * tilec, int *src, int *dst)
{
	int i;
	int *a = NULL;
	int *p = NULL;
	int w, l, h;
	int rw;			// width of the current resolution level computed                                                           */
	int rh;			// height of the current resolution level computed                                                          */
	int alter;
	int slice_w;

	w = tilec->x1-tilec->x0;
	h = tilec->y1-tilec->y0;
	l = tilec->numresolutions-1;

	// LL subband (upleft:   (0,rw) ~ (0, rh))
	rw = tilec->resolutions[0].x1 - tilec->resolutions[0].x0;
	rh = tilec->resolutions[0].y1 - tilec->resolutions[0].y0;
	alter = 0;
	slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : 1;
	slice_w = 1;
	a = src;
	p =	dst;
	mdc_reset_sb(a, p, rw, rh, w, h, LL_SPLIT, slice_w, alter);

	for (i = 1; i <= l; i++) {

		rw = tilec->resolutions[i - 1].x1 - tilec->resolutions[i - 1].x0;
		rh = tilec->resolutions[i - 1].y1 - tilec->resolutions[i - 1].y0;

		alter = ALTER_SPLIT ? (i % 2) : 0;
		slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : rw/2 << (i - 1);

		// HL subband (upright:   (0,rw) ~ (rh,2*rw-1))
		a = src + rw;
		p =	dst + rw;
		mdc_reset_sb(a, p, rw, rh, w, h, HL_SPLIT, slice_w, alter);

		// LH subband (downleft:  (rh,0) ~ (2*rh-1,rw))
		a = src + w * rh;
		p =	dst + w * rh;
		mdc_reset_sb(a, p, rw, rh, w, h, LH_SPLIT, slice_w, 1-alter);

		// HH subband (downright: (rh,rw) ~ (2*rh-1,2*rw-1))
		a = src + w * rh + rw;
		p =	dst + w * rh + rw;
		mdc_reset_sb(a, p, rw, rh, w, h, HH_SPLIT, slice_w, 1-alter);
	}
}

/* <summary>                                 */
/* generate prediction of one subband        */
/* </summary>                                */
void mdc_prediction_sb(int * src, int * dst, int rw, int rh, int w, int h, int opt, int slice_w, int lv, int sb_idx)
{
	int x, y;
	int i;
	int sw;
	int l, r;

	int scale0, scale1, scale2, scale3, scale;

	scale0 = 1;
	scale1 = 1;
	scale2 = 1;
	scale3 = 1;
	scale = 1 << (lv+3);

	sw = w * slice_w;

	switch(opt)
	{
	case PRED_HOR:
		for (y = 0; y < rh; y++)
		{
			for (x = 0; x < slice_w; x++)    // first col slice
			{
				l = x % slice_w;
				r = slice_w - l;
				*(dst + x) = *(src + x + r);
			}
			for (x = slice_w; x < rw - slice_w; x++)
			{
				l = x % slice_w;
				r = slice_w - l;
//				*(dst + x) = ( *(src + x - l - 1) + *(src + x + r) + 1 ) >> 1;
				*(dst + x) = ( *(src + x - l - 1) * r + *(src + x + r) * (l + 1) ) / (slice_w + 1);
			}
			for (x = rw - slice_w; x < rw; x++)    // last col slice
			{
				l = x % slice_w;
				r = slice_w - l;
				*(dst + x) = *(src + x - l - 1);
			}
			dst += w;
			src += w;
		}
		break;
	case PRED_VER:
		for (y = 0; y < slice_w; y++)    // first row slice
		{
			memcpy(dst, src + sw, sizeof(int)*rw);
			dst += w;
		}
		src += sw;
		for (y = slice_w; y < rh - slice_w; y+=slice_w)
		{
			for (i = 0; i < slice_w; i++)
			{
				for (x = 0; x < rw; x++)
				{
//					*(dst + x) = ( *(src - w + x) + *(src + sw + x) + 1 ) >> 1;
					*(dst + x) = ( *(src - w + x) * (slice_w - i) + *(src + sw + x) * (i + 1) ) / (slice_w + 1);
				}
				dst += w;
			}
			src += sw;
		}
		for (y = rh - slice_w; y < rh; y++)    // last row slice
		{
			memcpy(dst, src - w, sizeof(int)*rw);
			dst += w;
		}
		break;
	case PRED_DL:
		// first row
		*dst = *(src + w);
		for (x = 1; x < rw; x++)
		{
			*(dst + x) = *(src + w + x - 1);
		}

		for (y = 1; y < rh - 1; y++)
		{
			dst += w;
			src += w;
			*dst = *(src - w + 1);
			for (x = 1; x < rw - 1; x++)
			{
				*(dst + x) = ( *(src - w + x + 1) + *(src + w + x - 1) + 1 ) >> 1;
			}
			*(dst + x) = *(src + w + x - 1);
		}

		// last row
		dst += w;
		src += w;
		for (x = 0; x < rw - 1; x++)
		{
			*(dst + x) = *(src - w + x + 1);
		}
		*(dst + x) = *(src + x - w);
		break;
	case PRED_DR:
		// first row
		for (x = 0; x < rw - 1; x++)
		{
			*(dst + x) = *(src + w + x + 1);
		}
		*(dst + x) = *(src + w + x);

		for (y = 1; y < rh - 1; y++)
		{
			dst += w;
			src += w;
			*dst = *(src + w + 1);
			for (x = 1; x < rw - 1; x++)
			{
				*(dst + x) = ( *(src - w + x - 1) + *(src + w + x + 1) + 1 ) >> 1;
			}
			*(dst + x) = *(src - w + x - 1);
		}

		// last row
		dst += w;
		src += w;
		*dst = *(src - w);
		for (x = 1; x < rw; x++)
		{
			*(dst + x) = *(src - w + x - 1);
		}
		break;
	case PRED_CROSS:
		// first row
		*dst = (*(src + 1) + *(src + w) + 1) >> 1;  //upleft corner
		for (x = 1; x < rw - 1; x++)
		{
			*(dst + x) = ((*(src + x - 1) + *(src + x + 1) + *(src + x + w)) * 2 + 3) / 6;
		}
		*(dst + x) = (*(src + x - 1) + *(src + x + w) + 1) >> 1;  //upright corner
		
		for (y = 1; y < rh - 1; y++)
		{
			dst += w;
			src += w;
			
			*dst = ((*(src - w) + *(src + w) + *(src + 1)) * 2 + 3) / 6;    // first col
			for (x = 1; x < rw - 1; x++)
			{
				*(dst + x) = (*(src + x - w) + *(src + x + w) + *(src + x - 1) + *(src + x + 1) + 2) >> 2;
			}
			*(dst + x) = ((*(src + x - w) + *(src + x + w) + *(src + x - 1)) * 2 + 3) / 6;    // last col
		}

		// last row
		dst += w;
		src += w;
		*dst = (*(src + 1) + *(src - w) + 1) >> 1;  //downleft corner
		for (x = 1; x < rw - 1; x++)
		{
			*(dst + x) = ((*(src + x - 1) + *(src + x + 1) + *(src + x - w)) * 2 + 3) / 6;
		}
		*(dst + x) = (*(src + x - 1) + *(src + x - w) + 1) >> 1;  //downright corner
		break;
	case PRED_BWD: // high level subband -> low level subband
		for (y = 0; y < rh; y += 2)
		{
			for (x = 0; x < rw; x += 2)
			{
				*(dst + x) = *(src + x/2) * scale0 / scale;
				*(dst + x + 1) = *(src + x/2) * scale2 / scale;
			}
			dst += w;
			for (x = 0; x < rw; x += 2)
			{
				*(dst + x) = *(src + x/2) * scale1 / scale;
				*(dst + x + 1) = *(src + x/2) * scale3 / scale;
			}
			dst += w;
			src += w;
		}
		break;
	default:
		// no prediction (PRED_NO)
		for (y = 0; y < rh; y++)
		{
			memset(dst, 0, sizeof(int)*rw);
			dst += w;
		}
		break;
	}
}

/* <summary>                            */
/* generate mdc prediction error        */
/* </summary>                           */
void mdc_prediction_tile(opj_tcd_tilecomp_t * tilec, int *src, int *dst, int *org)
{
	int i;
	int *a = NULL;
	int *p = NULL;
	int *e = NULL;
	int w, l, h;
	int rw;			// width of the current resolution level computed                                                           */
	int rh;			// height of the current resolution level computed                                                          */
	int slice_w;
	int hl_pred, lh_pred, hh_pred;

	w = tilec->x1-tilec->x0;
	h = tilec->y1-tilec->y0;
	l = tilec->numresolutions-1;
	e = tilec->err_data;

	// LL subband (upright:   (0,rw) ~ (0,rh))
	rw = tilec->resolutions[0].x1 - tilec->resolutions[0].x0;
	rh = tilec->resolutions[0].y1 - tilec->resolutions[0].y0;
	a = src;
	p =	dst;
	slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : 1;
	slice_w = 1;
	mdc_prediction_sb(a, p, rw, rh, w, h, LL_PRED, slice_w, 0, LL_SUBBAND);

	// generate predictions
	for (i = 1; i <= l; i++) {
		if (i == 1)
		{
			hl_pred = NO_PRED;
			lh_pred = NO_PRED;
			hh_pred = NO_PRED;
		}
		else
		{
			hl_pred = HL_PRED;
			lh_pred = LH_PRED;
			hh_pred = HH_PRED;
		}

		rw = tilec->resolutions[i - 1].x1 - tilec->resolutions[i - 1].x0;
		rh = tilec->resolutions[i - 1].y1 - tilec->resolutions[i - 1].y0;

		slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : rw/2 << (i - 1);
		//slice_w = 1;

		// HL subband (upright:   (0,rw) ~ (rh,2*rw-1))
		// use horizontal slice / vertical prediction
		if (hl_pred == PRED_BWD)
		{
			a = src + rw/2;
			p =	dst + rw;
		}
		else
		{
			a = src + rw;
			p =	dst + rw;
		}
		mdc_prediction_sb(a, p, rw, rh, w, h, hl_pred, slice_w, i, HL_SUBBAND);

		// LH subband (downleft:  (rh,0) ~ (2*rh-1,rw))
		// use vertical slice / horizontal prediction
		if (lh_pred == PRED_BWD)
		{
			a = src + w * rh/2;
			p =	dst + w * rh;
		}
		else
		{
			a = src + w * rh;
			p =	dst + w * rh;
		}
		mdc_prediction_sb(a, p, rw, rh, w, h, lh_pred, slice_w, i, LH_SUBBAND);

		// HH subband (downright: (rh,rw) ~ (2*rh-1,2*rw-1))
		if (hh_pred == PRED_BWD)
		{
			a = src + w * rh/2 + rw/2;
			p =	dst + w * rh + rw;
		}
		else
		{
			a = src + w * rh + rw;
			p =	dst + w * rh + rw;
		}	
		mdc_prediction_sb(a, p, rw, rh, w, h, hh_pred, slice_w, i, HH_SUBBAND);
	}

	// generate prediction error
	a = org;
	p =	dst;
	for (i = 0; i < w * h; i ++)
	{
		*(e ++) = *(a ++) - *(p ++);
	}
}

/* <summary>                                */
/* generate reconstruction of one subband   */
/* </summary>                               */
void mdc_recon_sb(int *src, int *dst, int rw, int rh, int w, int h, int opt, int slice_w, int des_idx)
{
	int x, y;
	int i;
	int *a = NULL;
	int *p = NULL;
	int *e = NULL;
	int dw, sw;
	int rw1;
	int rh1;
	int rhs;

	dw = (w << 1);
	sw = (w * slice_w);

	switch(opt)
	{
	case SPLIT_VER:
		rw1 = rw / 2;
		rh1 = rh;
		if (des_idx == 0)    // description 1
		{
			a = dst + slice_w;
			p = src + slice_w;
		}
		else if (des_idx == 1)    // description 2
		{
			a = dst;
			p = src;
		}
		for (y = 0; y < rh1; y ++)
		{
			for (x = 0; x < rw1; x += slice_w)
			{
				for (i = 0; i < slice_w; i ++)
				{
					*(a + (x << 1) + i) += *(p + (x << 1) + i);
				}
			}
			a += w;
			p += w;
		}
		break;
	case SPLIT_HOR:
		rw1 = rw;
		rh1 = rh / 2;
		rhs = rh / 2 / slice_w * slice_w;
		if (des_idx == 0)    // description 1
		{
			a = dst + sw;
			p =	src + sw;
		}
		else if (des_idx == 1)    // description 2
		{
			a =	dst;
			p =	src;
		}
		for (y = 0; y < rhs; y += slice_w)
		{
			for (i = 0; i < slice_w; i ++)
			{
				for (x = 0; x < rw1; x ++)
				{
					*(a + x) += *(p + x);
				}
				a += w;
				p += w;
			}
			a += sw;
			p += sw;
		}

		if (des_idx == 0)
		{
			a = a - sw + (rh1 - y)*w;
			p = p - sw + (rh1 - y)*w;
		}
		for (i = 0; i < rh1 - y; i ++)
		{
			for (x = 0; x < rw1; x ++)
			{
				*(a + x) += *(p + x);
			}
			a += w;
			p += w;
		}
		break;
	case SPLIT_CB_U:
	case SPLIT_CB_L:
		rw1 = rw / 2;
		rh1 = rh / 2;
		a =	dst;
		p =	src;
		for (y = 0; y < rh1; y++)
		{
			for (x = 0; x < rw1; x++)
			{
				*(a + (x << 1) + 1 - des_idx) += *(p + (x << 1) + 1 - des_idx);
				*(a + (x << 1) + w + des_idx) += *(p + (x << 1) + w + des_idx);
			}
			a += dw;
			p += dw;
		}
		break;
	default:
		// NO_SPLIT;
		break;
	}
}

/* <summary>                                */
/* generate reconstruction of one tile      */
/* </summary>                               */
void mdc_recon_tile(opj_tcd_tilecomp_t * tilec, int *src, int *dst, int des_idx)
{
	int i;
	int *a = NULL;
	int *p = NULL;
	int w, l, h;
	int rw;			// width of the current resolution level computed                                                           */
	int rh;			// height of the current resolution level computed                                                          */
	int slice_w;

	w = tilec->x1-tilec->x0;
	h = tilec->y1-tilec->y0;
	l = tilec->numresolutions-1;

	// LL subband (upright:   (0,rw) ~ (0,rh))
	rw = tilec->resolutions[0].x1 - tilec->resolutions[0].x0;
	rh = tilec->resolutions[0].y1 - tilec->resolutions[0].y0;
	slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : 1;
	slice_w = 1;
	a = dst;
	p =	src;
	mdc_recon_sb(p, a, rw, rh, w, h, LL_SPLIT, slice_w, des_idx);

	// generate predictions
	for (i = 1; i <= l; i++) {
		
		rw = tilec->resolutions[i - 1].x1 - tilec->resolutions[i - 1].x0;
		rh = tilec->resolutions[i - 1].y1 - tilec->resolutions[i - 1].y0;

		des_idx = ALTER_SPLIT ? 1 - des_idx : des_idx;
		slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : rw/2 << (i - 1);

		// HL subband (upright:   (0,rw) ~ (rh,2*rw-1))
		// use horizontal slice / vertical prediction
		a = dst + rw;
		p =	src + rw;
		mdc_recon_sb(p, a, rw, rh, w, h, HL_SPLIT, slice_w, des_idx);

		// LH subband (downleft:  (rh,0) ~ (2*rh-1,rw))
		// use vertical slice / horizontal prediction
		a = dst + w * rh;
		p =	src + w * rh;
		mdc_recon_sb(p, a, rw, rh, w, h, LH_SPLIT, slice_w, 1-des_idx);

		// HH subband (downright: (rh,rw) ~ (2*rh-1,2*rw-1))
		a = dst + w * rh + rw;
		p =	src + w * rh + rw;
		mdc_recon_sb(p, a, rw, rh, w, h, HH_SPLIT, slice_w, 1-des_idx);

	}
}

/* <summary>                                     */
/* combine multiple descriptions for one subband */
/* </summary>                                    */
void mdc_combine_sb(int *src1, int *src2, int *dst, int rw, int rh, int w, int h, int opt, int slice_w, int alter, int sb_idx)
{
	int x, y;
	int *a = NULL;
	int *d1 = NULL;
	int *d2 = NULL;
	int dw, sw;
	int rw1;
	int rh1;
	int rhs;

	dw = (w << 1);
	sw = (w * slice_w);

	switch(opt)
	{
	case SPLIT_VER:
		rw1 = rw / 2;
		rh1 = rh;
		a = dst;
		if (alter)
		{
			d1 = src2;				// description 1
			d2 = src1 + slice_w;    // description 2
		}
		else
		{
			d1 = src1;				// description 1
			d2 = src2 + slice_w;    // description 2
		}
		for (y = 0; y < rh1; y ++)
		{
			for (x = 0; x < rw1; x += slice_w)
			{
				memcpy(a + (x << 1),           d1 + (x << 1), sizeof(int)*slice_w);
				memcpy(a + (x << 1) + slice_w, d2 + (x << 1), sizeof(int)*slice_w);
			}
			a += w;
			d1 += w;
			d2 += w;
		}
		break;
	case SPLIT_HOR:
		rw1 = rw;
		rh1 = rh / 2;
		rhs = rh / 2 / slice_w * slice_w;
		a = dst;
		if (alter)
		{
			d1 = src2;         // description 1
			d2 = src1 + sw;    // description 2
		}
		else
		{
			d1 = src1;         // description 1
			d2 = src2 + sw;    // description 2
		}
		for (y = 0; y < rhs; y += slice_w)
		{
			for (x = 0; x < slice_w; x ++)
			{
				memcpy(a, d1, sizeof(int)*rw1);
				a += w;
				d1 += w;
			}
			for (x = 0; x < slice_w; x ++)
			{
				memcpy(a, d2, sizeof(int)*rw1);
				a += w;
				d2 += w;
			}
			d1 += sw;
			d2 += sw;
		}

		d2 = d2 - sw + (rh1 - y)*w;
		for (x = 0; x < rh1 - y; x ++)
		{
			memcpy(a, d1, sizeof(int)*rw1);
			a += w;
			d1 += w;
		}
		for (x = 0; x < rh1 - y; x ++)
		{
			memcpy(a, d2, sizeof(int)*rw1);
			a += w;
			d2 += w;
		}
		break;
	case SPLIT_CB_U:
	case SPLIT_CB_L:
		rw1 = rw / 2;
		rh1 = rh / 2;
		a = dst;
		if (alter)
		{
			d1 = src2;        // description 1
			d2 = src1;        // description 2
		}
		else
		{
			d1 = src1;        // description 1
			d2 = src2;        // description 2
		}
		for (y = 0; y < rh / slice_w; y += 2)
		{
			for (x = 0; x < rw / slice_w; x += 2)
			{
				int i, j;

				for (i = 0; i < slice_w; i ++)
				{
					for (j = 0; j < slice_w; j ++)
					{
						*(a + j +  x    * slice_w + ( y    * slice_w + i) * w) = *(d1 + j +  x    * slice_w + ( y      * slice_w + i) * w);
						*(a + j +  x    * slice_w + ((y+1) * slice_w + i) * w) = *(d2 + j +  x    * slice_w + ((y + 1) * slice_w + i) * w);
						*(a + j + (x+1) * slice_w + ( y    * slice_w + i) * w) = *(d2 + j + (x+1) * slice_w + ( y      * slice_w + i) * w);
						*(a + j + (x+1) * slice_w + ((y+1) * slice_w + i) * w) = *(d1 + j + (x+1) * slice_w + ((y + 1) * slice_w + i) * w);
					}
				}
			}
		}
		break;
	default:
		// NO_SPLIT
		a = dst;

		if (sb_idx == LH_SUBBAND)
		{
			rh1 = rh;
			rw1 = rw / 2;

			d1 = src1;        // description 1
    		d2 = src2 + rw1;        // description 2

			for (y = 0; y < rh1; y++)
			{
				memcpy(a, d1, sizeof(int)*rw1);
				memcpy(a+rw1, d2, sizeof(int)*rw1);
				d1 += w;
				a += w;
				d2 += w;
			}
		}
		else
		{
			rh1 = rh / 2;

			d1 = src1;        // description 1
    		d2 = src2 + w*rh1;        // description 2

			for (y = 0; y < rh1; y++)
			{
				memcpy(a, d1, sizeof(int)*rw);
				a += w;
				d1 += w;
			}
			for (y = 0; y < rh1; y++)
			{
				memcpy(a, d2, sizeof(int)*rw);
				a += w;
				d2 += w;
			}
		}
		break;
	}
}

/* <summary>                                     */
/* combine multiple descriptions for one tile    */
/* </summary>                                    */
void mdc_combine_tile(opj_tcd_tilecomp_t * tilec, int *src1, int *src2, int *dst)
{
	int i;
	int *a = NULL;
	int *d1 = NULL;
	int *d2 = NULL;
	int w, l, h;
	int rw;			// width of the current resolution level computed                                                           */
	int rh;			// height of the current resolution level computed                                                          */
	int alter;
	int slice_w;

	w = tilec->x1-tilec->x0;
	h = tilec->y1-tilec->y0;
	l = tilec->numresolutions-1;

	// LL subband (upright:   (0,rw) ~ (0,rh))
	rw = tilec->resolutions[0].x1 - tilec->resolutions[0].x0;
	rh = tilec->resolutions[0].y1 - tilec->resolutions[0].y0;
	a = dst;
	d1 = src1;
	d2 = src2;
	alter = 0;
	slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : 1;
	slice_w = 1;
	mdc_combine_sb(d1, d2, a, rw, rh, w, h, LL_SPLIT, slice_w, alter, LL_SUBBAND);

	// generate predictions
	for (i = 1; i <= l; i++) {

		rw = tilec->resolutions[i - 1].x1 - tilec->resolutions[i - 1].x0;
		rh = tilec->resolutions[i - 1].y1 - tilec->resolutions[i - 1].y0;

		alter = ALTER_SPLIT ? (i % 2) : 0;
		slice_w = ALTER_SLICE ? min(ALTER_SLICE, rw/2) : rw/2 << (i - 1);

		// HL subband (upright:   (0,rw) ~ (rh,2*rw-1))
		// use horizontal slice / vertical prediction
		a = dst + rw;
		d1 = src1 + rw;
		d2 = src2 + rw;
		mdc_combine_sb(d1, d2, a, rw, rh, w, h, HL_SPLIT, slice_w, alter, HL_SUBBAND);

		// LH subband (downleft:  (rh,0) ~ (2*rh-1,rw))
		// use vertical slice / horizontal prediction
		a = dst + w * rh;
		d1 = src1 + w * rh;
		d2 = src2 + w * rh;
		mdc_combine_sb(d1, d2, a, rw, rh, w, h, LH_SPLIT, slice_w, 1-alter, LH_SUBBAND);

		// HH subband (downright: (rh,rw) ~ (2*rh-1,2*rw-1))
		a = dst + w * rh + rw;
		d1 = src1 + w * rh + rw;
		d2 = src2 + w * rh + rw;
		mdc_combine_sb(d1, d2, a, rw, rh, w, h, HH_SPLIT, slice_w, 1-alter, HH_SUBBAND);

	}
}

/* <summary>                                         */
/* decoder function                                  */
/* generate side decoded images                      */
/* </summary>                                        */
void mdc_side_decoder(opj_tcd_tilecomp_t * tilec, int des_idx)
{
    mdc_reset_tile(tilec, tilec->data, tilec->mdc_data[des_idx]);
	tilec->data = tilec->mdc_data[des_idx];

	mdc_prediction_tile(tilec, tilec->data, tilec->pred_data, tilec->data);

	mdc_recon_tile(tilec, tilec->pred_data, tilec->data, des_idx);
}

/* <summary>                                         */
/* decoder function                                  */
/* generate central decoded image                    */
/* </summary>                                        */
void mdc_central_decoder(opj_tcd_tilecomp_t * tilec)
{

	tilec->data = tilec->mdc_data[2];
    mdc_combine_tile(tilec, tilec->mdc_data[0], tilec->mdc_data[1], tilec->data);
}
